@@ -1,6 +1,25 @@
 // vim: tw=80
-use super::*;
-use syn::parse::{Parse, ParseStream};
+use proc_macro2::{Span, TokenStream};
+use quote::{format_ident, quote};
+use syn::{
+    *,
+    parse::{Parse, ParseStream},
+    spanned::Spanned
+};
+
+use crate::{
+    Attrs,
+    compile_error,
+    deanonymize,
+    deimplify,
+    demutify,
+    deselfify,
+    deselfify_args,
+    dewhereselfify,
+    find_ident_from_path,
+    gen_mock_ident,
+    AttrFormatter,
+};
 
 /// Make any implicit lifetime parameters explicit
 fn add_lifetime_parameters(sig: &mut Signature) {
@@ -35,13 +54,13 @@ fn add_lifetime_parameters(sig: &mut Signature) {
                     format_ident!("dont_care")
                 }
             };
-            let s = format!("'__mockall_{}", arg_ident);
+            let s = format!("'__mockall_{arg_ident}");
             let span = Span::call_site();
             let lt = Lifetime::new(&s, span);
             to.bounds.push(TypeParamBound::Lifetime(lt.clone()));
             generics.lt_token.get_or_insert(Token![<](span));
             generics.gt_token.get_or_insert(Token![>](span));
-            let gpl = GenericParam::Lifetime(LifetimeDef::new(lt));
+            let gpl = GenericParam::Lifetime(LifetimeParam::new(lt));
             generics.params.push(gpl);
         }
     }
@@ -94,12 +113,12 @@ fn add_lifetime_parameters(sig: &mut Signature) {
 
 /// Generate a #[derive(Debug)] Attribute
 fn derive_debug() -> Attribute {
+    let ml = parse2(quote!(derive(Debug))).unwrap();
     Attribute {
         pound_token: <Token![#]>::default(),
         style: AttrStyle::Outer,
         bracket_token: token::Bracket::default(),
-        path: Path::from(format_ident!("derive")),
-        tokens: quote!((Debug))
+        meta: Meta::List(ml)
     }
 }
 
@@ -128,7 +147,7 @@ fn mockable_item_impl(mut impl_: ItemImpl, name: &Ident, generics: &Generics)
 {
     mock_ident_in_type(&mut impl_.self_ty);
     for item in impl_.items.iter_mut() {
-        if let ImplItem::Method(ref mut iim) = item {
+        if let ImplItem::Fn(ref mut iim) = item {
             mockable_method(iim, name, generics);
         }
     }
@@ -136,12 +155,13 @@ fn mockable_item_impl(mut impl_: ItemImpl, name: &Ident, generics: &Generics)
 }
 
 /// Performs transformations on the method to make it mockable
-fn mockable_method(meth: &mut ImplItemMethod, name: &Ident, generics: &Generics)
+fn mockable_method(meth: &mut ImplItemFn, name: &Ident, generics: &Generics)
 {
     demutify(&mut meth.sig.inputs);
     deselfify_args(&mut meth.sig.inputs, name, generics);
     add_lifetime_parameters(&mut meth.sig);
     deimplify(&mut meth.sig.output);
+    dewhereselfify(&mut meth.sig.generics);
     if let ReturnType::Type(_, ty) = &mut meth.sig.output {
         deselfify(ty, name, generics);
         deanonymize(ty);
@@ -151,7 +171,7 @@ fn mockable_method(meth: &mut ImplItemMethod, name: &Ident, generics: &Generics)
 
 /// Performs transformations on the method to make it mockable
 fn mockable_trait_method(
-    meth: &mut TraitItemMethod,
+    meth: &mut TraitItemFn,
     name: &Ident,
     generics: &Generics)
 {
@@ -159,6 +179,7 @@ fn mockable_trait_method(
     deselfify_args(&mut meth.sig.inputs, name, generics);
     add_lifetime_parameters(&mut meth.sig);
     deimplify(&mut meth.sig.output);
+    dewhereselfify(&mut meth.sig.generics);
     if let ReturnType::Type(_, ty) = &mut meth.sig.output {
         deselfify(ty, name, generics);
         deanonymize(ty);
@@ -173,9 +194,9 @@ fn mockable_trait(trait_: ItemTrait, name: &Ident, generics: &Generics)
     let items = trait_.items.into_iter()
     .map(|ti| {
         match ti {
-            TraitItem::Method(mut tim) => {
-                mockable_trait_method(&mut tim, name, generics);
-                ImplItem::Method(tim2iim(tim, &Visibility::Inherited))
+            TraitItem::Fn(mut tif) => {
+                mockable_trait_method(&mut tif, name, generics);
+                ImplItem::Fn(tif2iif(tif, &Visibility::Inherited))
             },
             TraitItem::Const(tic) => {
                 ImplItem::Const(tic2iic(tic, &Visibility::Inherited))
@@ -246,6 +267,7 @@ fn tic2iic(tic: TraitItemConst, vis: &syn::Visibility) -> ImplItemConst {
         vis: vis.clone(),
         defaultness: None,
         const_token: tic.const_token,
+        generics: tic.generics,
         ident: tic.ident,
         colon_token: tic.colon_token,
         ty: tic.ty,
@@ -255,15 +277,15 @@ fn tic2iic(tic: TraitItemConst, vis: &syn::Visibility) -> ImplItemConst {
     }
 }
 
-/// Converts a TraitItemMethod into an ImplItemMethod
-fn tim2iim(m: syn::TraitItemMethod, vis: &syn::Visibility)
-    -> syn::ImplItemMethod
+/// Converts a TraitItemFn into an ImplItemFn
+fn tif2iif(m: syn::TraitItemFn, vis: &syn::Visibility)
+    -> syn::ImplItemFn
 {
     let empty_block = Block {
         brace_token: token::Brace::default(),
         stmts: Vec::new()
     };
-    syn::ImplItemMethod{
+    syn::ImplItemFn{
         attrs: m.attrs,
         vis: vis.clone(),
         defaultness: None,
@@ -293,40 +315,48 @@ fn tit2iit(tit: TraitItemType, vis: &Visibility) -> ImplItemType {
     }
 }
 
+/// Like a TraitItemFn, but with a visibility
+struct TraitItemVFn {
+    pub vis: Visibility,
+    pub tif: TraitItemFn
+}
+
+impl Parse for TraitItemVFn {
+    fn parse(input: ParseStream) -> syn::parse::Result<Self> {
+        let attrs = input.call(Attribute::parse_outer)?;
+        let vis: syn::Visibility = input.parse()?;
+        let mut tif: TraitItemFn = input.parse()?;
+        tif.attrs = attrs;
+        Ok(Self{vis, tif})
+    }
+}
+
 pub(crate) struct MockableStruct {
     pub attrs: Vec<Attribute>,
     pub consts: Vec<ImplItemConst>,
     pub generics: Generics,
     /// Inherent methods of the mockable struct
-    pub methods: Vec<ImplItemMethod>,
+    pub methods: Vec<ImplItemFn>,
     pub name: Ident,
     pub vis: Visibility,
-    pub impls: Vec<ItemImpl>
+    pub impls: Vec<ItemImpl>,
 }
 
 impl MockableStruct {
     /// Does this struct derive Debug?
     pub fn derives_debug(&self) -> bool {
         self.attrs.iter()
-        .any(|attr| {
-            if let Ok(Meta::List(ml)) = attr.parse_meta() {
-                let i = ml.path.get_ident();
-                if i.map_or(false, |i| *i == "derive") {
-                    ml.nested.iter()
-                    .any(|nm| {
-                        if let NestedMeta::Meta(m) = nm {
-                            let i = m.path().get_ident();
-                            i.map_or(false, |i| *i == "Debug")
-                        } else {
-                            false
-                        }
-                    })
-                } else {
-                    false
-                }
-            } else {
-                false
+        .any(|attr|{
+            let mut derive_debug = false;
+            if attr.path().is_ident("derive") {
+                attr.parse_nested_meta(|meta| {
+                    if meta.path.is_ident("Debug") {
+                        derive_debug = true;
+                    }
+                    Ok(())
+                }).unwrap();
             }
+            derive_debug
         })
     }
 }
@@ -334,7 +364,16 @@ impl MockableStruct {
 impl From<(Attrs, ItemTrait)> for MockableStruct {
     fn from((attrs, item_trait): (Attrs, ItemTrait)) -> MockableStruct {
         let trait_ = attrs.substitute_trait(&item_trait);
-        let mut attrs = trait_.attrs.clone();
+        // Strip "must_use" from a trait definition.  For traits, the "must_use"
+        // should apply only when the trait is used like "impl Trait" or "dyn
+        // Trait".  So it shouldn't necessarily affect the mock struct that
+        // implements the trait.
+        let mut attrs = AttrFormatter::new(&trait_.attrs)
+            .doc(true)
+            .async_trait(true)
+            .trait_variant(true)
+            .must_use(false)
+            .format();
         attrs.push(derive_debug());
         let vis = trait_.vis.clone();
         let name = gen_mock_ident(&trait_.ident);
@@ -357,6 +396,17 @@ impl From<ItemImpl> for MockableStruct {
         let name = match &*item_impl.self_ty {
             Type::Path(type_path) => {
                 let n = find_ident_from_path(&type_path.path).0;
+                let self_generics = &type_path.path.segments.last().unwrap().arguments;
+                if let PathArguments::AngleBracketed(abga) = &self_generics {
+                    if item_impl.generics.params.len() != abga.args.len() {
+                        // If a struct's impl block has elided lifetimes, then
+                        // they won't show up in the impl block's generics
+                        // list.  automock can't currently handle that.
+                        // https://github.com/asomers/mockall/issues/610
+                        compile_error(item_impl.span(),
+                            "automock does not currently support structs with elided lifetimes");
+                    }
+                }
                 gen_mock_ident(&n)
             },
             x => {
@@ -370,8 +420,7 @@ impl From<ItemImpl> for MockableStruct {
         let mut consts = Vec::new();
         let generics = item_impl.generics.clone();
         let mut methods = Vec::new();
-        let pub_token = Token![pub](Span::call_site());
-        let vis = Visibility::Public(VisPublic{pub_token});
+        let vis = Visibility::Public(Token![pub](Span::call_site()));
         let mut impls = Vec::new();
         if let Some((bang, _path, _)) = &item_impl.trait_ {
             if bang.is_some() {
@@ -386,7 +435,7 @@ impl From<ItemImpl> for MockableStruct {
                 match item {
                     ImplItem::Const(_iic) =>
                         (),
-                    ImplItem::Method(_meth) =>
+                    ImplItem::Fn(_meth) =>
                         (),
                     ImplItem::Type(ty) => {
                         attrs.attrs.insert(ty.ident.clone(), ty.ty.clone());
@@ -399,7 +448,7 @@ impl From<ItemImpl> for MockableStruct {
         } else {
             for item in item_impl.items.into_iter() {
                 match item {
-                    ImplItem::Method(mut meth) => {
+                    ImplItem::Fn(mut meth) => {
                         mockable_method(&mut meth, &name, &item_impl.generics);
                         methods.push(meth)
                     },
@@ -438,10 +487,12 @@ impl Parse for MockableStruct {
         while !impl_content.is_empty() {
             let item: ImplItem = impl_content.parse()?;
             match item {
-                ImplItem::Method(mut iim) => {
+                ImplItem::Verbatim(ts) => {
+                    let tivf: TraitItemVFn = parse2(ts)?;
+                    let mut iim = tif2iif(tivf.tif, &tivf.vis);
                     mockable_method(&mut iim, &name, &generics);
                     methods.push(iim);
-                },
+                }
                 ImplItem::Const(iic) => consts.push(iic),
                 _ => {
                     return Err(input.error("Unsupported in this context"));
@@ -453,20 +504,18 @@ impl Parse for MockableStruct {
         while !input.is_empty() {
             let item: Item = input.parse()?;
             match item {
-                Item::Trait(it) => {
-                    let note = "Deprecated mock! syntax.  Instead of \"trait X\", write \"impl X for Y\".  See PR #205";
-                    let mut impl_ = mockable_trait(it, &name, &generics);
-                    impl_.attrs.push(Attribute {
-                        pound_token: <token::Pound>::default(),
-                        style: AttrStyle::Outer,
-                        bracket_token: token::Bracket::default(),
-                        path: Path::from(format_ident!("deprecated")),
-                        tokens: quote!((since = "0.9.0", note = #note)),
-                    });
-                    impls.push(impl_)
-                },
-                Item::Impl(ii) =>
-                    impls.push(mockable_item_impl(ii, &name, &generics)),
+                Item::Impl(mut ii) => {
+                    for item in ii.items.iter_mut() {
+                        // Convert any methods that syn couldn't parse as
+                        // ImplItemFn.
+                        if let ImplItem::Verbatim(ts) = item {
+                            let tif: TraitItemFn = parse2(ts.clone()).unwrap();
+                            let iim = tif2iif(tif, &Visibility::Inherited);
+                            *item = ImplItem::Fn(iim);
+                        }
+                    }
+                    impls.push(mockable_item_impl(ii, &name, &generics));
+                }
                 _ => return Err(input.error("Unsupported in this context")),
             }
         }
@@ -494,7 +543,7 @@ mod add_lifetime_parameters {
 
     #[test]
     fn array() {
-        let mut meth: TraitItemMethod = parse2(quote!(
+        let mut meth: TraitItemFn = parse2(quote!(
             fn foo(&self, x: [&dyn T; 1]);
         )).unwrap();
         add_lifetime_parameters(&mut meth.sig);
@@ -507,7 +556,7 @@ mod add_lifetime_parameters {
 
     #[test]
     fn bare_fn_with_named_args() {
-        let mut meth: TraitItemMethod = parse2(quote!(
+        let mut meth: TraitItemFn = parse2(quote!(
             fn foo(&self, x: fn(&dyn T));
         )).unwrap();
         add_lifetime_parameters(&mut meth.sig);
@@ -519,7 +568,7 @@ mod add_lifetime_parameters {
 
     #[test]
     fn plain() {
-        let mut meth: TraitItemMethod = parse2(quote!(
+        let mut meth: TraitItemFn = parse2(quote!(
             fn foo(&self, x: &dyn T);
         )).unwrap();
         add_lifetime_parameters(&mut meth.sig);
@@ -532,7 +581,7 @@ mod add_lifetime_parameters {
 
     #[test]
     fn slice() {
-        let mut meth: TraitItemMethod = parse2(quote!(
+        let mut meth: TraitItemFn = parse2(quote!(
             fn foo(&self, x: &[&dyn T]);
         )).unwrap();
         add_lifetime_parameters(&mut meth.sig);
@@ -545,7 +594,7 @@ mod add_lifetime_parameters {
 
     #[test]
     fn tuple() {
-        let mut meth: TraitItemMethod = parse2(quote!(
+        let mut meth: TraitItemFn = parse2(quote!(
             fn foo(&self, x: (&dyn T, u32));
         )).unwrap();
         add_lifetime_parameters(&mut meth.sig);
@@ -558,7 +607,7 @@ mod add_lifetime_parameters {
 
     #[test]
     fn with_anonymous_lifetime() {
-        let mut meth: TraitItemMethod = parse2(quote!(
+        let mut meth: TraitItemFn = parse2(quote!(
             fn foo(&self, x: &(dyn T + '_));
         )).unwrap();
         add_lifetime_parameters(&mut meth.sig);
@@ -570,7 +619,7 @@ mod add_lifetime_parameters {
 
     #[test]
     fn with_parens() {
-        let mut meth: TraitItemMethod = parse2(quote!(
+        let mut meth: TraitItemFn = parse2(quote!(
             fn foo(&self, x: &(dyn T));
         )).unwrap();
         add_lifetime_parameters(&mut meth.sig);
@@ -583,7 +632,7 @@ mod add_lifetime_parameters {
 
     #[test]
     fn with_lifetime_parameter() {
-        let mut meth: TraitItemMethod = parse2(quote!(
+        let mut meth: TraitItemFn = parse2(quote!(
             fn foo<'a>(&self, x: &(dyn T + 'a));
         )).unwrap();
         add_lifetime_parameters(&mut meth.sig);
@@ -595,7 +644,7 @@ mod add_lifetime_parameters {
 
     #[test]
     fn with_static_lifetime() {
-        let mut meth: TraitItemMethod = parse2(quote!(
+        let mut meth: TraitItemFn = parse2(quote!(
             fn foo(&self, x: &(dyn T + 'static));
         )).unwrap();
         add_lifetime_parameters(&mut meth.sig);
@@ -613,8 +662,8 @@ mod sanity_check_sig {
     #[test]
     #[should_panic(expected = "Mockall does not support \"impl trait\" in argument position.  Use \"T: SomeTrait\" instead.")]
     fn impl_trait() {
-        let meth: ImplItemMethod = parse2(quote!(
-            fn foo(&self, x: impl SomeTrait);
+        let meth: ImplItemFn = parse2(quote!(
+            fn foo(&self, x: impl SomeTrait) {}
         )).unwrap();
         sanity_check_sig(&meth.sig);
     }
